"""
awslimitchecker/services/ec2.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import abc  # noqa
import os
import logging
from collections import defaultdict
from copy import deepcopy

import botocore

from .base import _AwsService
from ..limit import AwsLimit

logger = logging.getLogger(__name__)

RI_NO_AZ = 'xxREGIONAL_BENEFIT-NO_AZxx'


class _Ec2Service(_AwsService):

    service_name = 'EC2'
    api_name = 'ec2'
    quotas_service_code = 'ec2'

    #: Mapping of lower-case instance family character (instance type first
    #: character) to limit name for that family.
    instance_family_to_limit_name = {
        'f': 'Running On-Demand All F instances',
        'g': 'Running On-Demand All G instances',
        'p': 'Running On-Demand All P instances',
        'x': 'Running On-Demand All X instances'
    }

    #: Mapping of lower-case instance family character to Service Quotas
    #: quota name for that family.
    instance_family_to_quota_name = {
        'f': 'Running On-Demand F instances',
        'g': 'Running On-Demand G instances',
        'p': 'Running On-Demand P instances',
        'x': 'Running On-Demand X instances'
    }

    #: Name of default limit for all other (standard) instance families.
    default_limit_name = 'Running On-Demand All Standard ' \
                         '(A, C, D, H, I, M, R, T, Z) instances'

    #: Name of default Service Quota for all other (standard) families.
    default_quota_name = 'Running On-Demand Standard ' \
                         '(A, C, D, H, I, M, R, T, Z) instances'

    #: List of instance types that aren't exposed via Service Quotas
    no_quotas_types = [
        'c5d.12xlarge',
        'c5d.24xlarge',
        'c5d.metal',
        'cc1.4xlarge',
        'cg1.4xlarge',
        'cr1.8xlarge',
        'g4dn.metal',
        'hi1.4xlarge',
        'hs1.8xlarge',
        'm5dn.metal',
        'm5n.metal',
        'r5dn.metal',
        'r5n.metal',
        'u-18tb1.metal',
        'u-24tb1.metal'
    ]

    def find_usage(self):
        """
        Determine the current usage for each limit of this service,
        and update corresponding Limit via
        :py:meth:`~.AwsLimit._add_current_usage`.
        """
        logger.debug("Checking usage for service %s", self.service_name)
        self.connect()
        self.connect_resource()
        for lim in self.limits.values():
            lim._reset_usage()
        if self._use_vcpu_limits:
            self._find_usage_instances_vcpu()
        else:
            self._find_usage_instances_nonvcpu()
        self._find_usage_networking_sgs()
        self._find_usage_networking_eips()
        self._find_usage_networking_eni_sg()
        self._find_usage_spot_instances()
        self._find_usage_spot_fleets()
        self._have_usage = True
        logger.debug("Done checking usage.")

    def _find_usage_instances_nonvcpu(self):
        """calculate On-Demand instance usage for all types and update Limits"""
        # update our limits with usage
        inst_usage = self._instance_usage()
        res_usage = self._get_reserved_instance_count()
        logger.debug('Reserved instance count: %s', res_usage)
        total_ris = 0
        running_ris = 0
        # subtract reservations from instance usage
        ondemand_usage = defaultdict(int)
        for az in inst_usage:
            if az not in res_usage:
                for i_type, count in inst_usage[az].items():
                    ondemand_usage[i_type] += count
                continue
            # else we have reservations for this AZ
            for i_type, count in inst_usage[az].items():
                if i_type not in res_usage[az]:
                    # no reservations for this type
                    ondemand_usage[i_type] += count
                    continue
                od = count - res_usage[az][i_type]
                total_ris += res_usage[az][i_type]
                if count < res_usage[az][i_type]:
                    running_ris += count
                else:
                    running_ris += res_usage[az][i_type]
                if od < 0:
                    # we have unused reservations
                    continue
                ondemand_usage[i_type] += od
        logger.debug(
            'Found %d total RIs and %d running/used RIs',
            total_ris, running_ris
        )
        total_instances = 0
        for i_type, usage in ondemand_usage.items():
            key = 'Running On-Demand {t} instances'.format(
                t=i_type)
            self.limits[key]._add_current_usage(
                usage,
                aws_type='AWS::EC2::Instance',
            )
            total_instances += usage
        # limit for ALL On-Demand EC2 instances
        key = 'Running On-Demand EC2 instances'
        self.limits[key]._add_current_usage(
            total_instances,
            aws_type='AWS::EC2::Instance'
        )

    def _find_usage_instances_vcpu(self):
        res_usage = self._get_reserved_instance_count()
        logger.debug('Reserved instance count: %s', res_usage)
        usage = self._instance_usage_vcpu(res_usage)
        limit_values = defaultdict(int)
        for i_family, count in usage.items():
            limname = self.instance_family_to_limit_name.get(
                i_family, self.default_limit_name
            )
            limit_values[limname] += count
        for lname in list(
            self.instance_family_to_limit_name.values()
        ) + [self.default_limit_name]:
            if lname not in limit_values:
                limit_values[lname] = 0
        for limname, count in limit_values.items():
            self.limits[limname]._add_current_usage(
                count,
                aws_type='AWS::EC2::Instance',
            )

    def _find_usage_spot_instances(self):
        """calculate spot instance request usage and update Limits"""
        logger.debug('Getting spot instance request usage')
        try:
            res = self.conn.describe_spot_instance_requests()
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == 'UnsupportedOperation':
                return
            raise
        count = 0
        for req in res['SpotInstanceRequests']:
            if req['State'] in ['open', 'active']:
                count += 1
                logger.debug('Counting spot instance request %s state=%s',
                             req['SpotInstanceRequestId'], req['State'])
            else:
                logger.debug('NOT counting spot instance request %s state=%s',
                             req['SpotInstanceRequestId'], req['State'])
        self.limits['Max spot instance requests per region']._add_current_usage(
            count,
            aws_type='AWS::EC2::SpotInstanceRequest'
        )

    def _find_usage_spot_fleets(self):
        """calculate spot fleet request usage and update Limits"""
        logger.debug('Getting spot fleet request usage')
        try:
            res = self.conn.describe_spot_fleet_requests()
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == 'UnsupportedOperation':
                return
            raise
        if 'NextToken' in res:
            logger.error('Error: describe_spot_fleet_requests() response '
                         'includes pagination token, but pagination not '
                         'configured in awslimitchecker.')
        active_fleets = 0
        total_target_cap = 0
        lim_cap_per_fleet = self.limits['Max target capacity per spot fleet']
        lim_launch_specs = self.limits[
            'Max launch specifications per spot fleet']
        for fleet in res['SpotFleetRequestConfigs']:
            _id = fleet['SpotFleetRequestId']
            if fleet['SpotFleetRequestState'] != 'active':
                logger.debug('Skipping spot fleet request %s in state %s',
                             _id, fleet['SpotFleetRequestState'])
                continue
            active_fleets += 1
            cap = fleet['SpotFleetRequestConfig']['TargetCapacity']
            launch_specs = len(
                fleet['SpotFleetRequestConfig']['LaunchSpecifications'])
            total_target_cap += cap
            lim_cap_per_fleet._add_current_usage(
                cap, resource_id=_id, aws_type='AWS::EC2::SpotFleetRequest')
            lim_launch_specs._add_current_usage(
                launch_specs, resource_id=_id,
                aws_type='AWS::EC2::SpotFleetRequest')
        self.limits['Max active spot fleets per region']._add_current_usage(
            active_fleets, aws_type='AWS::EC2::SpotFleetRequest'
        )
        self.limits['Max target capacity for all spot '
                    'fleets in region']._add_current_usage(
            total_target_cap, aws_type='AWS::EC2::SpotFleetRequest'
        )

    def _get_reserved_instance_count(self):
        """
        For each availability zone, get the count of current instance
        reservations of each instance type. Return as a nested
        dict of AZ name to dict of instance type to reservation count.

        :rtype: dict
        """
        reservations = defaultdict(int)
        az_to_res = {}
        logger.debug("Getting reserved instance information")
        res = self.conn.describe_reserved_instances()

        for x in res['ReservedInstances']:
            if x['State'] != 'active':
                logger.debug("Skipping ReservedInstance %s with state %s",
                             x['ReservedInstancesId'], x['State'])
                continue
            if 'AvailabilityZone' not in x:
                # "Regional Benefit" AZ-less reservation
                x['AvailabilityZone'] = RI_NO_AZ
            if x['AvailabilityZone'] not in az_to_res:
                az_to_res[x['AvailabilityZone']] = deepcopy(reservations)
            az_to_res[x['AvailabilityZone']][
                x['InstanceType']] += x['InstanceCount']
        # flatten and return
        for x in az_to_res:
            az_to_res[x] = dict(az_to_res[x])
        return az_to_res

    def _instance_usage(self):
        """
        Find counts of currently-running EC2 Instances
        (On-Demand or Reserved) by placement (Availability
        Zone) and instance type (size). Return as a nested dict
        of AZ name to dict of instance type to count.

        :rtype: dict
        """
        # On-Demand instances by type
        ondemand = {}
        for t in self._instance_types():
            ondemand[t] = 0
        az_to_inst = {}
        logger.debug("Getting usage for on-demand instances")
        for inst in self.resource_conn.instances.all():
            if inst.spot_instance_request_id:
                logger.info("Spot instance found (%s); skipping from "
                            "Running On-Demand Instances count", inst.id)
                continue
            if inst.placement.get('Tenancy', 'default') != 'default':
                logger.info(
                    'Skipping instance %s with Tenancy %s',
                    inst.id, inst.placement['Tenancy']
                )
                continue
            if inst.state['Name'] in ['stopped', 'terminated']:
                logger.debug("Ignoring instance %s in state %s", inst.id,
                             inst.state['Name'])
                continue
            if inst.placement['AvailabilityZone'] not in az_to_inst:
                az_to_inst[
                    inst.placement['AvailabilityZone']] = deepcopy(ondemand)
            try:
                az_to_inst[
                    inst.placement['AvailabilityZone']][inst.instance_type] += 1
            except KeyError:
                logger.error("ERROR - unknown instance type '%s'; not "
                             "counting", inst.instance_type)
        return az_to_inst

    def _instance_usage_vcpu(self, ris):
        """
        Find counts of currently-running EC2 Instance vCPUs
        (On-Demand or Reserved) by instance family. Return as a dict of
        instance family letter to count.

        :param ris: nested dict of reserved instances, as returned by
          :py:meth:`~._get_reserved_instance_count`
        :type ris: dict
        :rtype: dict
        """
        inst_counts = defaultdict(int)
        logger.debug("Getting usage for on-demand instances (vCPU limit)")
        for inst in self.resource_conn.instances.all():
            if inst.spot_instance_request_id:
                logger.info("Spot instance found (%s); skipping from "
                            "Running On-Demand Instances count", inst.id)
                continue
            if inst.placement.get('Tenancy', 'default') != 'default':
                logger.info(
                    'Skipping instance %s with Tenancy %s',
                    inst.id, inst.placement['Tenancy']
                )
                continue
            if inst.state['Name'] in ['stopped', 'terminated']:
                logger.debug("Ignoring instance %s in state %s", inst.id,
                             inst.state['Name'])
                continue
            az = inst.placement['AvailabilityZone']
            itype = inst.instance_type
            if ris.get(az, {}).get(itype, 0) > 0:
                logger.debug(
                    'Using RI for %s: %s in %s', inst.id, itype, az
                )
                ris[az][itype] -= 1
                continue
            inst_counts[inst.instance_type[0]] += (
                inst.cpu_options['CoreCount'] * inst.cpu_options[
                    'ThreadsPerCore'
                ]
            )
        return inst_counts

    @property
    def _use_vcpu_limits(self):
        """
        Return whether or not to use the new vCPU-based limits.

        :return: whether to use vCPU-based limits (True) or older
          per-instance-type limits (False)
        :rtype: bool
        """
        if 'USE_VCPU_LIMITS' in os.environ:
            if os.environ['USE_VCPU_LIMITS'] == 'true':
                logger.debug(
                    'Using vCPU-based EC2 limits due to USE_VCPU_LIMITS=true '
                    'in environment.'
                )
                return True
            logger.debug(
                'Using vCPU-based EC2 limits due to USE_VCPU_LIMITS in '
                'environment and set to something other than "true".'
            )
            return False
        oldconn = self.conn
        self.connect()
        region_name = self.conn._client_config.region_name
        self.conn = oldconn
        if region_name.startswith('cn-') or region_name.startswith('us-gov-'):
            logger.debug(
                'Using non-vCPU EC2 limits due to region name: %s', region_name
            )
            return False
        return True

    def get_limits(self):
        """
        Return all known limits for this service, as a dict of their names
        to :py:class:`~.AwsLimit` objects.

        :returns: dict of limit names to :py:class:`~.AwsLimit` objects
        :rtype: dict
        """
        if self.limits != {}:
            return self.limits
        limits = {}
        if self._use_vcpu_limits:
            limits.update(self._get_limits_instances_vcpu())
        else:
            limits.update(self._get_limits_instances_nonvcpu())
        limits.update(self._get_limits_networking())
        limits.update(self._get_limits_spot())
        self.limits = limits
        return self.limits

    def _update_limits_from_api(self):
        """
        Query EC2's DescribeAccountAttributes API action, and update limits
        with the quotas returned. Updates ``self.limits``.
        """
        self.connect()
        self.connect_resource()
        logger.info("Querying EC2 DescribeAccountAttributes for limits")
        # no need to paginate
        attribs = self.conn.describe_account_attributes()
        for attrib in attribs['AccountAttributes']:
            aname = attrib['AttributeName']
            val = attrib['AttributeValues'][0]['AttributeValue']
            lname = None
            if aname == 'max-elastic-ips':
                lname = 'Elastic IP addresses (EIPs)'
            elif aname == 'max-instances':
                lname = 'Running On-Demand EC2 instances'
            elif aname == 'vpc-max-elastic-ips':
                lname = 'VPC Elastic IP addresses (EIPs)'
            elif aname == 'vpc-max-security-groups-per-interface':
                lname = 'VPC security groups per elastic network interface'
            if lname in self.limits:
                if int(val) == 0:
                    continue
                self.limits[lname]._set_api_limit(int(val))
        logger.debug("Done setting limits from API")

    def _get_limits_instances_nonvcpu(self):
        """
        Return a dict of limits for EC2 instances only, for regions using
        non-vCPU-based (old-style) On Demand Instances limits.
        This method should only be used internally by
        :py:meth:~.get_limits`.

        :rtype: dict
        """
        # from: http://aws.amazon.com/ec2/faqs/
        # (On-Demand, Reserved, Spot)
        default_limits = (20, 20, 5)
        special_limits = {
            'c4.4xlarge': (10, 20, 5),
            'c4.8xlarge': (5, 20, 5),
            'c5.4xlarge': (10, 20, 5),
            'c5.9xlarge': (5, 20, 5),
            'c5.18xlarge': (5, 20, 5),
            'cg1.4xlarge': (2, 20, 5),
            'cr1.8xlarge': (2, 20, 5),
            'd2.4xlarge': (10, 20, 5),
            'd2.8xlarge': (5, 20, 5),
            'g2.2xlarge': (5, 20, 5),
            'g2.8xlarge': (2, 20, 5),
            'g3.4xlarge': (1, 20, 5),
            'g3.8xlarge': (1, 20, 5),
            'g3.16xlarge': (1, 20, 5),
            'h1.8xlarge': (10, 20, 5),
            'h1.16xlarge': (5, 20, 5),
            'hi1.4xlarge': (2, 20, 5),
            'hs1.8xlarge': (2, 20, 0),
            'i2.2xlarge': (8, 20, 0),
            'i2.4xlarge': (4, 20, 0),
            'i2.8xlarge': (2, 20, 0),
            'i2.xlarge': (8, 20, 0),
            'i3.2xlarge': (2, 20, 0),
            'i3.4xlarge': (2, 20, 0),
            'i3.8xlarge': (2, 20, 0),
            'i3.16xlarge': (2, 20, 0),
            'i3.large': (2, 20, 0),
            'i3.xlarge': (2, 20, 0),
            'm4.4xlarge': (10, 20, 5),
            'm4.10xlarge': (5, 20, 5),
            'm4.16xlarge': (5, 20, 5),
            'm5.4xlarge': (10, 20, 5),
            'm5.12xlarge': (5, 20, 5),
            'm5.24xlarge': (5, 20, 5),
            'p2.8xlarge': (1, 20, 5),
            'p2.16xlarge': (1, 20, 5),
            'p2.xlarge': (1, 20, 5),
            'p3.2xlarge': (1, 20, 5),
            'p3.8xlarge': (1, 20, 5),
            'p3.16xlarge': (1, 20, 5),
            'p3dn.24xlarge': (1, 20, 5),
            'r3.4xlarge': (10, 20, 5),
            'r3.8xlarge': (5, 20, 5),
            'r4.4xlarge': (10, 20, 5),
            'r4.8xlarge': (5, 20, 5),
            'r4.16xlarge': (1, 20, 5),
        }
        limits = {}
        for i_type in self._instance_types():
            key = 'Running On-Demand {t} instances'.format(
                t=i_type)
            lim = default_limits[0]
            if i_type in special_limits:
                lim = special_limits[i_type][0]
            quotas_name = 'Running On-Demand %s instances' % i_type
            if i_type in self.no_quotas_types:
                quotas_name = None
            limits[key] = AwsLimit(
                key,
                self,
                lim,
                self.warning_threshold,
                self.critical_threshold,
                limit_type='On-Demand instances',
                limit_subtype=i_type,
                ta_limit_name='On-Demand instances - %s' % i_type,
                quotas_name=quotas_name
            )
        # limit for ALL running On-Demand instances
        key = 'Running On-Demand EC2 instances'
        limits[key] = AwsLimit(
            key,
            self,
            default_limits[0],
            self.warning_threshold,
            self.critical_threshold,
            limit_type='On-Demand instances',
            quotas_name='Total running On-Demand instances'
        )
        return limits

    def _get_limits_instances_vcpu(self):
        """
        Return a dict of limits for EC2 instances only, for regions using
        vCPU-based (new-style) On Demand Instances limits.
        This method should only be used internally by
        :py:meth:~.get_limits`.

        :rtype: dict
        """
        limits = {}
        iftln = self.instance_family_to_limit_name
        for key in iftln.keys():
            limits[iftln[key]] = AwsLimit(
                iftln[key],
                self,
                128,
                self.warning_threshold,
                self.critical_threshold,
                limit_type='On-Demand instances',
                limit_subtype=key.upper(),
                quotas_name=self.instance_family_to_quota_name[key]
            )
        limits[self.default_limit_name] = AwsLimit(
            self.default_limit_name,
            self,
            1152,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='On-Demand instances',
            limit_subtype='Standard',
            quotas_name=self.default_quota_name
        )
        return limits

    def _get_limits_spot(self):
        """
        Return a dict of limits for spot requests only.
        This method should only be used internally by
        :py:meth:~.get_limits`.

        :rtype: dict
        """
        limits = {}
        limits['Max spot instance requests per region'] = AwsLimit(
            'Max spot instance requests per region',
            self,
            20,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='Spot instance requests'
        )

        limits['Max active spot fleets per region'] = AwsLimit(
            'Max active spot fleets per region',
            self,
            1000,
            self.warning_threshold,
            self.critical_threshold,
        )

        limits['Max launch specifications per spot fleet'] = AwsLimit(
            'Max launch specifications per spot fleet',
            self,
            50,
            self.warning_threshold,
            self.critical_threshold,
        )

        limits['Max target capacity per spot fleet'] = AwsLimit(
            'Max target capacity per spot fleet',
            self,
            3000,
            self.warning_threshold,
            self.critical_threshold
        )

        limits['Max target capacity for all spot fleets in region'] = AwsLimit(
            'Max target capacity for all spot fleets in region',
            self,
            5000,
            self.warning_threshold,
            self.critical_threshold
        )
        return limits

    def _find_usage_networking_sgs(self):
        """calculate usage for VPC-related things"""
        logger.debug("Getting usage for EC2 VPC resources")
        sg_count = 0
        rules_per_sg = defaultdict(int)
        for sg in self.resource_conn.security_groups.all():
            if sg.vpc_id is None:
                continue
            sg_count += 1
            """
            see: https://github.com/jantman/awslimitchecker/issues/431

            The value for each of ingress and egress is the count of all
            PrefixListIds in all rules, plus the count of all
            UserIdGroupPairs in all rules, plus the maximum of:
              the count of all IpRanges in all rules
                 -or-
              the count of all Ipv6Ranges in all rules

            The limit that we alert on is the maximum of those values for
            ingress and egress.

            In short, behind the scenes, there are four firewall rulesets
            per SG: (IPv4|IPv6) (ingress|egress)
            Each can have a maximum of <limit> entries. PrefixListIds and
            UserIdGroupPairs count towards both IPv4 and IPv6.
            """
            counts = []
            for perm in [sg.ip_permissions, sg.ip_permissions_egress]:
                counts.append(
                    max(
                        sum([len(x.get('IpRanges', [])) for x in perm]),
                        sum([len(x.get('Ipv6Ranges', [])) for x in perm])
                    ) +
                    sum([len(x.get('PrefixListIds', [])) for x in perm]) +
                    sum([len(x.get('UserIdGroupPairs', [])) for x in perm])
                )
            rules_per_sg[sg.id] = max(counts)
        # set usage
        self.limits['VPC security groups per Region']._add_current_usage(
            sg_count,
            aws_type='AWS::EC2::SecurityGroup',
        )
        for sg_id, count in rules_per_sg.items():
            self.limits['Rules per VPC security group']._add_current_usage(
                count,
                aws_type='AWS::EC2::SecurityGroupRule',
                resource_id=sg_id,
            )

    def _find_usage_networking_eips(self):
        logger.debug("Getting usage for EC2 EIPs")
        vpc_addrs = self.resource_conn.vpc_addresses.all()
        self.limits['VPC Elastic IP addresses (EIPs)']._add_current_usage(
            sum(1 for a in vpc_addrs if a.domain == 'vpc'),
            aws_type='AWS::EC2::EIP',
        )
        # the EC2 limits screen calls this 'EC2-Classic Elastic IPs'
        # but Trusted Advisor just calls it 'Elastic IP addresses (EIPs)'
        classic_addrs = self.resource_conn.classic_addresses.all()
        self.limits['Elastic IP addresses (EIPs)']._add_current_usage(
            sum(1 for a in classic_addrs if a.domain == 'standard'),
            aws_type='AWS::EC2::EIP',
        )

    def _find_usage_networking_eni_sg(self):
        logger.debug("Getting usage for EC2 Network Interfaces")
        ints = self.resource_conn.network_interfaces.all()
        for iface in ints:
            if iface.vpc is None:
                continue
            self.limits[
                'VPC security groups per elastic network interface'
            ]._add_current_usage(
                len(iface.groups),
                aws_type='AWS::EC2::NetworkInterface',
                resource_id=iface.id,
            )

    def _get_limits_networking(self):
        """
        Return a dict of VPC-related limits only.
        This method should only be used internally by
        :py:meth:~.get_limits`.

        :rtype: dict
        """
        limits = {}
        limits['VPC security groups per Region'] = AwsLimit(
            'VPC security groups per Region',
            self,
            2500,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::EC2::SecurityGroup',
            limit_subtype='AWS::EC2::VPC',
            quotas_name='VPC security groups per Region',
            quotas_service_code='vpc'
        )
        limits['Rules per VPC security group'] = AwsLimit(
            'Rules per VPC security group',
            self,
            60,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::EC2::SecurityGroup',
            limit_subtype='AWS::EC2::VPC',
            quotas_name='Inbound or outbound rules per security group',
            quotas_service_code='vpc'
        )
        limits['VPC Elastic IP addresses (EIPs)'] = AwsLimit(
            'VPC Elastic IP addresses (EIPs)',
            self,
            5,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::EC2::EIP',
            limit_subtype='AWS::EC2::VPC',
            ta_service_name='VPC',  # TA shows this as VPC not EC2
            quotas_name='Number of EIPs - VPC EIPs'
        )
        # the EC2 limits screen calls this 'EC2-Classic Elastic IPs'
        # but Trusted Advisor just calls it 'Elastic IP addresses (EIPs)'
        limits['Elastic IP addresses (EIPs)'] = AwsLimit(
            'Elastic IP addresses (EIPs)',
            self,
            5,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::EC2::EIP',
            quotas_name='Elastic IP addresses for EC2-Classic'
        )
        limits['VPC security groups per elastic network interface'] = AwsLimit(
            'VPC security groups per elastic network interface',
            self,
            5,
            self.warning_threshold,
            self.critical_threshold,
            limit_type='AWS::EC2::SecurityGroup',
            limit_subtype='AWS::EC2::NetworkInterface',
        )
        return limits

    def required_iam_permissions(self):
        """
        Return a list of IAM Actions required for this Service to function
        properly. All Actions will be shown with an Effect of "Allow"
        and a Resource of "*".

        :returns: list of IAM Action strings
        :rtype: list
        """
        return [
            "ec2:DescribeAccountAttributes",
            "ec2:DescribeAddresses",
            "ec2:DescribeInstances",
            "ec2:DescribeInternetGateways",
            "ec2:DescribeNetworkAcls",
            "ec2:DescribeNetworkInterfaces",
            "ec2:DescribeReservedInstances",
            "ec2:DescribeRouteTables",
            "ec2:DescribeSecurityGroups",
            "ec2:DescribeSnapshots",
            "ec2:DescribeSpotDatafeedSubscription",
            "ec2:DescribeSpotFleetInstances",
            "ec2:DescribeSpotFleetRequestHistory",
            "ec2:DescribeSpotFleetRequests",
            "ec2:DescribeSpotInstanceRequests",
            "ec2:DescribeSpotPriceHistory",
            "ec2:DescribeSubnets",
            "ec2:DescribeVolumes",
            "ec2:DescribeVpcs",
        ]

    def _instance_types(self):
        """
        Return a list of all known EC2 instance types

        :returns: list of all valid known EC2 instance types
        :rtype: list
        """
        GENERAL_TYPES = [
            'a1.2xlarge',
            'a1.4xlarge',
            'a1.large',
            'a1.medium',
            'a1.metal',
            'a1.xlarge',
            'm3.2xlarge',
            'm3.large',
            'm3.medium',
            'm3.xlarge',
            'm4.2xlarge',
            'm4.4xlarge',
            'm4.10xlarge',
            'm4.16xlarge',
            'm4.large',
            'm4.xlarge',
            'm5.2xlarge',
            'm5.4xlarge',
            'm5.8xlarge',
            'm5.12xlarge',
            'm5.16xlarge',
            'm5.24xlarge',
            'm5.large',
            'm5.metal',
            'm5.xlarge',
            'm5a.2xlarge',
            'm5a.4xlarge',
            'm5a.8xlarge',
            'm5a.12xlarge',
            'm5a.16xlarge',
            'm5a.24xlarge',
            'm5a.large',
            'm5a.xlarge',
            'm5ad.2xlarge',
            'm5ad.4xlarge',
            'm5ad.8xlarge',
            'm5ad.12xlarge',
            'm5ad.16xlarge',
            'm5ad.24xlarge',
            'm5ad.large',
            'm5ad.xlarge',
            'm5d.2xlarge',
            'm5d.4xlarge',
            'm5d.8xlarge',
            'm5d.12xlarge',
            'm5d.16xlarge',
            'm5d.24xlarge',
            'm5d.large',
            'm5d.metal',
            'm5d.xlarge',
            'm5dn.2xlarge',
            'm5dn.4xlarge',
            'm5dn.8xlarge',
            'm5dn.12xlarge',
            'm5dn.16xlarge',
            'm5dn.24xlarge',
            'm5dn.large',
            'm5dn.metal',
            'm5dn.xlarge',
            'm5n.2xlarge',
            'm5n.4xlarge',
            'm5n.8xlarge',
            'm5n.12xlarge',
            'm5n.16xlarge',
            'm5n.24xlarge',
            'm5n.large',
            'm5n.metal',
            'm5n.xlarge',
            't2.2xlarge',
            't2.large',
            't2.medium',
            't2.micro',
            't2.nano',
            't2.small',
            't2.xlarge',
            't3.2xlarge',
            't3.large',
            't3.medium',
            't3.micro',
            't3.nano',
            't3.small',
            't3.xlarge',
            't3a.2xlarge',
            't3a.large',
            't3a.medium',
            't3a.micro',
            't3a.nano',
            't3a.small',
            't3a.xlarge',
        ]

        PREV_GENERAL_TYPES = [
            't1.micro',
            'm1.small',
            'm1.medium',
            'm1.large',
            'm1.xlarge',
        ]

        MEMORY_TYPES = [
            'r3.2xlarge',
            'r3.4xlarge',
            'r3.8xlarge',
            'r3.large',
            'r3.xlarge',
            'r4.2xlarge',
            'r4.4xlarge',
            'r4.8xlarge',
            'r4.16xlarge',
            'r4.large',
            'r4.xlarge',
            'r5.2xlarge',
            'r5.4xlarge',
            'r5.8xlarge',
            'r5.12xlarge',
            'r5.16xlarge',
            'r5.24xlarge',
            'r5.large',
            'r5.metal',
            'r5.xlarge',
            'r5a.2xlarge',
            'r5a.4xlarge',
            'r5a.8xlarge',
            'r5a.12xlarge',
            'r5a.16xlarge',
            'r5a.24xlarge',
            'r5a.large',
            'r5a.xlarge',
            'r5ad.2xlarge',
            'r5ad.4xlarge',
            'r5ad.8xlarge',
            'r5ad.12xlarge',
            'r5ad.16xlarge',
            'r5ad.24xlarge',
            'r5ad.large',
            'r5ad.xlarge',
            'r5d.2xlarge',
            'r5d.4xlarge',
            'r5d.8xlarge',
            'r5d.12xlarge',
            'r5d.16xlarge',
            'r5d.24xlarge',
            'r5d.large',
            'r5d.metal',
            'r5d.xlarge',
            'r5dn.2xlarge',
            'r5dn.4xlarge',
            'r5dn.8xlarge',
            'r5dn.12xlarge',
            'r5dn.16xlarge',
            'r5dn.24xlarge',
            'r5dn.large',
            'r5dn.metal',
            'r5dn.xlarge',
            'r5n.2xlarge',
            'r5n.4xlarge',
            'r5n.8xlarge',
            'r5n.12xlarge',
            'r5n.16xlarge',
            'r5n.24xlarge',
            'r5n.large',
            'r5n.metal',
            'r5n.xlarge',
            'u-18tb1.metal',
            'u-24tb1.metal',
            'x1.16xlarge',
            'x1.32xlarge',
            'x1e.2xlarge',
            'x1e.4xlarge',
            'x1e.8xlarge',
            'x1e.16xlarge',
            'x1e.32xlarge',
            'x1e.xlarge',
            'z1d.2xlarge',
            'z1d.3xlarge',
            'z1d.6xlarge',
            'z1d.12xlarge',
            'z1d.large',
            'z1d.xlarge',
        ]

        PREV_MEMORY_TYPES = [
            'm2.xlarge',
            'm2.2xlarge',
            'm2.4xlarge',
            'cr1.8xlarge',
        ]

        COMPUTE_TYPES = [
            'c3.2xlarge',
            'c3.4xlarge',
            'c3.8xlarge',
            'c3.large',
            'c3.xlarge',
            'c4.2xlarge',
            'c4.4xlarge',
            'c4.8xlarge',
            'c4.large',
            'c4.xlarge',
            'c5.2xlarge',
            'c5.4xlarge',
            'c5.9xlarge',
            'c5.12xlarge',
            'c5.18xlarge',
            'c5.24xlarge',
            'c5.large',
            'c5.metal',
            'c5.xlarge',
            'c5d.2xlarge',
            'c5d.4xlarge',
            'c5d.9xlarge',
            'c5d.12xlarge',
            'c5d.18xlarge',
            'c5d.24xlarge',
            'c5d.large',
            'c5d.metal',
            'c5d.xlarge',
            'c5n.2xlarge',
            'c5n.4xlarge',
            'c5n.9xlarge',
            'c5n.18xlarge',
            'c5n.large',
            'c5n.metal',
            'c5n.xlarge',
        ]

        PREV_COMPUTE_TYPES = [
            'c1.medium',
            'c1.xlarge',
            'cc2.8xlarge',
            'cc1.4xlarge',
        ]

        ACCELERATED_COMPUTE_TYPES = [
            'f1.4xlarge',
            'p2.xlarge',
            'p2.8xlarge',
            'p2.16xlarge',
            'p3.16xlarge',
            'p3.2xlarge',
            'p3.8xlarge',
            'p3dn.24xlarge',
        ]

        STORAGE_TYPES = [
            'h1.2xlarge',
            'h1.4xlarge',
            'h1.8xlarge',
            'h1.16xlarge',
            'i2.2xlarge',
            'i2.4xlarge',
            'i2.8xlarge',
            'i2.xlarge',
            'i3.2xlarge',
            'i3.4xlarge',
            'i3.8xlarge',
            'i3.16xlarge',
            'i3.large',
            'i3.metal',
            'i3.xlarge',
            'i3en.2xlarge',
            'i3en.3xlarge',
            'i3en.6xlarge',
            'i3en.12xlarge',
            'i3en.24xlarge',
            'i3en.large',
            'i3en.xlarge',
        ]

        PREV_STORAGE_TYPES = [
            # NOTE hi1.4xlarge is no longer in the instance type listings,
            # but some accounts might still have a limit for it
            'hi1.4xlarge',
            'hs1.8xlarge',
        ]

        DENSE_STORAGE_TYPES = [
            'd2.xlarge',
            'd2.2xlarge',
            'd2.4xlarge',
            'd2.8xlarge',
        ]

        GPU_TYPES = [
            'g2.2xlarge',
            'g2.8xlarge',
            'g3.4xlarge',
            'g3.8xlarge',
            'g3.16xlarge',
            'g3s.xlarge',
            'g4dn.2xlarge',
            'g4dn.4xlarge',
            'g4dn.8xlarge',
            'g4dn.12xlarge',
            'g4dn.16xlarge',
            'g4dn.metal',
            'g4dn.xlarge',
        ]

        PREV_GPU_TYPES = [
            'cg1.4xlarge',
        ]

        FPGA_TYPES = [
            # note, as of 2016-12-17, these are still in Developer Preview;
            # there isn't a published instance limit yet, so we'll assume
            # it's the default...
            'f1.2xlarge',
            'f1.16xlarge',
        ]

        return (
            GENERAL_TYPES +
            PREV_GENERAL_TYPES +
            MEMORY_TYPES +
            PREV_MEMORY_TYPES +
            COMPUTE_TYPES +
            PREV_COMPUTE_TYPES +
            ACCELERATED_COMPUTE_TYPES +
            STORAGE_TYPES +
            PREV_STORAGE_TYPES +
            DENSE_STORAGE_TYPES +
            GPU_TYPES +
            PREV_GPU_TYPES +
            FPGA_TYPES
        )



"""
awslimitchecker/quotas.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

##############################################################################
Copyright 2015-2019 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
##############################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
##############################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
##############################################################################
"""

from botocore.exceptions import ClientError
import logging

from awslimitchecker.connectable import Connectable

logger = logging.getLogger(__name__)


class ServiceQuotasClient(Connectable):
    api_name = 'service-quotas'

    def __init__(self, boto_connection_kwargs):
        """
        Client for the AWS Service Quotas service, that manages retrieving
        quotas information and updating :py:class:`~.AwsLimit` instances for
        them. This class is also intended to cache Service Quotas responses.

        :param boto_connection_kwargs: keyword arguments to pass to boto3
          connection methods.
        :type boto_connection_kwargs: dict
        """
        self._boto3_connection_kwargs = boto_connection_kwargs
        self._cache = {}
        self.conn = None

    def quotas_for_service(self, service_code):
        """
        Return this account's current quotas for the specified service code.
        Also cache them on this class instance.

        :param service_code: the service code to get quotas for
        :type service_code: str
        :return: QuotaName to dictionary of quota information returned by the
          service
        :rtype: dict
        """
        if service_code in self._cache:
            return self._cache[service_code]
        self.connect()
        logger.debug(
            'Getting service quotas for service code: %s', service_code
        )
        self._cache[service_code] = {}
        try:
            paginator = self.conn.get_paginator('list_service_quotas')
            for page in paginator.paginate(ServiceCode=service_code):
                for item in page['Quotas']:
                    if item['QuotaName'] in self._cache[service_code]:
                        logger.error(
                            'ERROR: Received duplicate service quota for '
                            'service code %s quota name "%s" - QuotaCodes %s'
                            ' and %s', service_code, item['QuotaName'],
                            self._cache[service_code][
                                item['QuotaName']
                            ]['QuotaCode'], item['QuotaCode']
                        )
                    self._cache[service_code][item['QuotaName'].lower()] = item
        except ClientError as ex:
            if ex.response.get(
                'Error', {}
            ).get('Code', '') == 'NoSuchResourceException':
                logger.warning(
                    'Attempted to retrieve Service Quotas for service code '
                    '%s but received NoSuchResourceException',
                    service_code
                )
                return {}
            raise
        logger.debug(
            'Retrieved %d quotas for service code %s: %s',
            len(self._cache[service_code]), service_code,
            sorted([x['QuotaName'] for x in self._cache[service_code].values()])
        )
        return self._cache[service_code]

    def get_quota_value(
        self, service_code, quota_name, units='None', converter=None
    ):
        """
        Return a given quota value, or None if it cannot be found. If
        ``units`` is a value other than ``None``, attempt to convert the value
        to the specified units.

        :param service_code: the service code to get a quota from
        :type service_code: str
        :param quota_name: the quota name to get
        :type quota_name: str
        :param units: the units for the value, or the string "None"
        :type units: str
        :param converter: A callable for unit conversion.
          Must take three positional arguments: the Service Quotas value for
          this quota (float), the quota ``Unit`` str, and the return value of
          :py:meth:`~.quotas_unit`. This callable is responsible for converting
          the quota value from the quota Unit to this class's expected unit.
          If they cannot be converted, it should log an error and return None.
        :type converter: ``callable``
        :return: the quota value
        :rtype: float or None
        """
        svc = self.quotas_for_service(service_code)
        if quota_name.lower() not in svc:
            return None
        val = svc[quota_name.lower()].get('Value', None)
        if svc[quota_name.lower()]['Unit'] != units:
            if converter is not None:
                return converter(val, svc[quota_name.lower()]['Unit'], units)
            logger.error(
                'ERROR: Service Quota service_code=%s QuotaName="%s" has '
                'Units set to "%s", but expected units to be "%s"; '
                'awslimitchecker does not know how to '
                'handle this. This quota will be ignored. Please open a bug '
                'report.', service_code, quota_name,
                svc[quota_name.lower()]['Unit'], units
            )
            return None
        return val



"""
awslimitchecker/tests/services/test_eks.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
from awslimitchecker.tests.services import result_fixtures
from awslimitchecker.services.eks import _EksService

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import patch, call, Mock, DEFAULT, ANY
else:
    from unittest.mock import patch, call, Mock, DEFAULT, ANY


pbm = 'awslimitchecker.services.eks'  # module patch base
pb = '%s._EksService' % pbm  # class patch pase


class Test_EksService(object):

    def test_init(self):
        """test __init__()"""
        cls = _EksService(21, 43, {}, None)
        assert cls.service_name == 'EKS'
        assert cls.api_name == 'eks'
        assert cls.conn is None
        assert cls.warning_threshold == 21
        assert cls.critical_threshold == 43

    def test_get_limits(self):
        cls = _EksService(21, 43, {}, None)
        cls.limits = {}
        res = cls.get_limits()
        assert sorted(res.keys()) == sorted([
            'Clusters',
            'Control plane security groups per cluster',
            'Managed node groups per cluster',
            'Nodes per managed node group',
            'Public endpoint access CIDR ranges per cluster',
            'Fargate profiles per cluster',
            'Selectors per Fargate profile',
            'Label pairs per Fargate profile selector',
        ])
        for name, limit in res.items():
            assert limit.service == cls
            assert limit.def_warning_threshold == 21
            assert limit.def_critical_threshold == 43

    def test_get_limits_again(self):
        """test that existing limits dict is returned on subsequent calls"""
        mock_limits = Mock()
        cls = _EksService(21, 43, {}, None)
        cls.limits = mock_limits
        res = cls.get_limits()
        assert res == mock_limits

    def test_find_usage(self):
        """test find usage method calls other methods"""
        mock_conn = Mock()
        with patch('%s.connect' % pb) as mock_connect:
            with patch.multiple(
                pb,
                _find_clusters_usage=DEFAULT,
            ) as mocks:
                cls = _EksService(21, 43, {}, None)
                cls.conn = mock_conn
                assert cls._have_usage is False
                cls.find_usage()
        assert mock_connect.mock_calls == [call()]
        assert cls._have_usage is True
        assert mock_conn.mock_calls == []
        for x in [
            '_find_clusters_usage',
        ]:
            assert mocks[x].mock_calls == [call()]

    def test_find_clusters_usage(self):
        list_clusters = result_fixtures.EKS.test_find_clusters_usage_list
        describe_cluster = result_fixtures.EKS.test_find_clusters_usage_describe
        list_nodegroups = result_fixtures.EKS.test_find_clusters_usage_nodegrps
        list_fargates = result_fixtures.EKS.test_find_clusters_usage_fargates
        dsc_fargate = result_fixtures.EKS.test_find_clusters_usage_fargate_prof

        clusters_limit_key = 'Clusters'
        security_group_limit_key = 'Control plane security groups per cluster'
        nodegroup_limit_key = 'Managed node groups per cluster'
        public_cidr_limit_key = 'Public endpoint access CIDR ranges per cluster'
        fargate_profiles_limit_key = 'Fargate profiles per cluster'
        selectors_limit_key = 'Selectors per Fargate profile'
        label_pairs_limit_key = 'Label pairs per Fargate profile selector'

        mock_conn = Mock()
        mock_conn.list_clusters.return_value = list_clusters
        mock_conn.describe_cluster.side_effect = describe_cluster
        mock_conn.list_nodegroups.side_effect = list_nodegroups
        mock_conn.list_fargate_profiles.side_effect = list_fargates
        mock_conn.describe_fargate_profile.side_effect = dsc_fargate

        cls = _EksService(21, 43, {'region_name': 'us-west-2'}, None)
        cls.conn = mock_conn
        cls._find_clusters_usage()

        assert mock_conn.mock_calls == [
            call.list_clusters(),
            call.describe_cluster(name=ANY),
            call.list_nodegroups(clusterName=ANY),
            call.list_fargate_profiles(clusterName=ANY),
            call.describe_fargate_profile(
                clusterName=ANY,
                fargateProfileName=ANY
            ),
            call.describe_cluster(name=ANY),
            call.list_nodegroups(clusterName=ANY),
            call.list_fargate_profiles(clusterName=ANY),
            call.describe_fargate_profile(
                clusterName=ANY,
                fargateProfileName=ANY
            ),
            call.describe_fargate_profile(
                clusterName=ANY,
                fargateProfileName=ANY
            ),
        ]
        assert len(cls.limits[clusters_limit_key].get_current_usage()) == 1
        assert cls.limits[clusters_limit_key].get_current_usage()[
            0].get_value() == 2

        assert len(cls.limits[
            security_group_limit_key].get_current_usage()) == 2
        assert cls.limits[security_group_limit_key].get_current_usage()[
            0].get_value() == 1
        assert cls.limits[security_group_limit_key].get_current_usage()[
            1].get_value() == 2

        assert len(cls.limits[
            nodegroup_limit_key].get_current_usage()) == 2
        assert cls.limits[nodegroup_limit_key].get_current_usage()[
            0].get_value() == 2
        assert cls.limits[nodegroup_limit_key].get_current_usage()[
            1].get_value() == 1

        assert len(cls.limits[
            public_cidr_limit_key].get_current_usage()) == 2
        assert cls.limits[public_cidr_limit_key].get_current_usage()[
            0].get_value() == 3
        assert cls.limits[public_cidr_limit_key].get_current_usage()[
            1].get_value() == 1

        assert len(cls.limits[
            fargate_profiles_limit_key].get_current_usage()) == 2
        assert cls.limits[fargate_profiles_limit_key].get_current_usage()[
            0].get_value() == 1
        assert cls.limits[fargate_profiles_limit_key].get_current_usage()[
            1].get_value() == 2

        assert len(cls.limits[
            selectors_limit_key].get_current_usage()) == 3
        assert cls.limits[selectors_limit_key].get_current_usage()[
            0].get_value() == 1
        assert cls.limits[selectors_limit_key].get_current_usage()[
            1].get_value() == 2
        assert cls.limits[selectors_limit_key].get_current_usage()[
            2].get_value() == 3

        assert len(cls.limits[
            label_pairs_limit_key].get_current_usage()) == 6
        assert cls.limits[label_pairs_limit_key].get_current_usage()[
            0].get_value() == 1
        assert cls.limits[label_pairs_limit_key].get_current_usage()[
            1].get_value() == 1
        assert cls.limits[label_pairs_limit_key].get_current_usage()[
            2].get_value() == 2
        assert cls.limits[label_pairs_limit_key].get_current_usage()[
            3].get_value() == 1
        assert cls.limits[label_pairs_limit_key].get_current_usage()[
            4].get_value() == 2
        assert cls.limits[label_pairs_limit_key].get_current_usage()[
            5].get_value() == 3

    def test_required_iam_permissions(self):
        cls = _EksService(21, 43, {}, None)
        assert cls.required_iam_permissions() == [
            "eks:ListClusters",
            "eks:DescribeCluster",
            "eks:ListNodegroups",
            "eks:ListFargateProfiles",
            "eks:DescribeFargateProfile"
        ]



"""
awslimitchecker/tests/alerts/test_pagerdutyv1.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2019 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
import json
from awslimitchecker.alerts.pagerdutyv1 import PagerDutyV1
import pytest

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import patch, call, Mock, DEFAULT
else:
    from unittest.mock import patch, call, Mock, DEFAULT

pbm = 'awslimitchecker.alerts.pagerdutyv1'
pb = '%s.PagerDutyV1' % pbm


class TestInit(object):

    @patch.dict(
        'os.environ',
        {'PAGERDUTY_SERVICE_KEY_CRIT': 'cKey'},
        clear=True
    )
    def test_no_options(self):
        cls = PagerDutyV1('foo')
        assert cls._region_name == 'foo'
        assert cls._service_key_crit == 'cKey'
        assert cls._service_key_warn is None
        assert cls._account_alias is None
        assert cls._incident_key == 'awslimitchecker--foo'

    @patch.dict(
        'os.environ',
        {
            'PAGERDUTY_SERVICE_KEY_CRIT': 'cKey',
            'PAGERDUTY_SERVICE_KEY_WARN': 'wKey'
        },
        clear=True
    )
    def test_warn_env_var(self):
        cls = PagerDutyV1('foo')
        assert cls._region_name == 'foo'
        assert cls._service_key_crit == 'cKey'
        assert cls._service_key_warn == 'wKey'
        assert cls._account_alias is None
        assert cls._incident_key == 'awslimitchecker--foo'

    @patch.dict('os.environ', {}, clear=True)
    def test_all_options(self):
        cls = PagerDutyV1(
            'foo', account_alias='myacct', critical_service_key='cKey',
            warning_service_key='wKey',
            incident_key='foo{account_alias}bar{region_name}baz'
        )
        assert cls._region_name == 'foo'
        assert cls._service_key_crit == 'cKey'
        assert cls._service_key_warn == 'wKey'
        assert cls._account_alias == 'myacct'
        assert cls._incident_key == 'foomyacctbarfoobaz'

    @patch.dict('os.environ', {}, clear=True)
    def test_no_crit_key(self):
        with pytest.raises(RuntimeError) as exc:
            PagerDutyV1('foo')
        assert str(exc.value) == 'ERROR: PagerDutyV1 alert ' \
                                 'provider requires ' \
                                 'critical_service_key parameter or' \
                                 ' PAGERDUTY_SERVICE_KEY_CRIT ' \
                                 'environment variable.'


class PagerDutyV1Tester(object):

    def setup(self):
        with patch('%s.__init__' % pb) as m_init:
            m_init.return_value = None
            self.cls = PagerDutyV1('rname')
            self.cls._incident_key = 'iKey'
            self.cls._region_name = 'rname'
            self.cls._service_key_warn = None
            self.cls._service_key_crit = None
            self.cls._account_alias = None


class TestSendEvent(PagerDutyV1Tester):

    def test_success(self):
        mock_http = Mock()
        mock_resp = Mock(
            status=200, data='{"status": "success", "message": '
                             '"Event processed", "incident_key":'
                             ' "iKey"}'
        )
        mock_http.request.return_value = mock_resp
        expected = json.dumps(
            {'foo': 'bar', 'service_key': 'sKey'}, sort_keys=True
        ).encode('utf-8')
        with patch('%s.urllib3.PoolManager' % pbm) as mock_pm:
            mock_pm.return_value = mock_http
            self.cls._send_event('sKey', {'foo': 'bar'})
        assert mock_http.mock_calls == [
            call.request(
                'POST', self.cls.pd_url,
                headers={'Content-type': 'application/json'},
                body=expected
            )
        ]

    def test_invalid_event(self):
        mock_http = Mock()
        mock_resp = Mock(
            status=400, data='{"status": "invalid event",'
                             '"message": "Event object is invalid", '
                             '"errors": ["foo"]}'
        )
        mock_http.request.return_value = mock_resp
        expected = json.dumps(
            {'foo': 'bar', 'service_key': 'sKey'}, sort_keys=True
        ).encode('utf-8')
        with patch('%s.urllib3.PoolManager' % pbm) as mock_pm:
            mock_pm.return_value = mock_http
            with pytest.raises(RuntimeError):
                self.cls._send_event('sKey', {'foo': 'bar'})
        assert mock_http.mock_calls == [
            call.request(
                'POST', self.cls.pd_url,
                headers={'Content-type': 'application/json'},
                body=expected
            )
        ]


class TestEventDict(PagerDutyV1Tester):

    def test_no_account_alias(self):
        assert self.cls._event_dict() == {
            'incident_key': 'iKey',
            'details': {
                'region': 'rname'
            },
            'client': 'awslimitchecker'
        }

    def test_with_account_alias(self):
        self.cls._account_alias = 'myAcct'
        assert self.cls._event_dict() == {
            'incident_key': 'iKey',
            'details': {
                'region': 'rname',
                'account_alias': 'myAcct'
            },
            'client': 'awslimitchecker'
        }


class TestOnSuccess(PagerDutyV1Tester):

    def test_happy_path(self):
        self.cls._service_key_crit = 'cKey'
        self.cls._service_key_warn = 'wKey'
        self.cls._account_alias = 'myAcct'
        with patch('%s._event_dict' % pb, autospec=True) as m_ed:
            m_ed.return_value = {'event': 'dict', 'details': {}}
            with patch('%s._send_event' % pb, autospec=True) as m_send:
                self.cls.on_success(duration=12.34567)
        assert m_ed.mock_calls == [call(self.cls)]
        assert m_send.mock_calls == [
            call(self.cls, 'cKey', {
                'event': 'dict',
                'details': {
                    'duration_seconds': 12.34567
                },
                'event_type': 'resolve',
                'description': 'awslimitchecker in myAcct rname found '
                               'no problems; run completed in 12.35 '
                               'seconds'
            }),
            call(self.cls, 'wKey', {
                'event': 'dict',
                'details': {
                    'duration_seconds': 12.34567
                },
                'event_type': 'resolve',
                'description': 'awslimitchecker in myAcct rname found '
                               'no problems; run completed in 12.35 '
                               'seconds'
            })
        ]

    def test_no_duration(self):
        self.cls._service_key_crit = 'cKey'
        self.cls._service_key_warn = 'wKey'
        self.cls._account_alias = 'myAcct'
        with patch('%s._event_dict' % pb, autospec=True) as m_ed:
            m_ed.return_value = {'event': 'dict', 'details': {}}
            with patch('%s._send_event' % pb, autospec=True) as m_send:
                self.cls.on_success()
        assert m_ed.mock_calls == [call(self.cls)]
        assert m_send.mock_calls == [
            call(self.cls, 'cKey', {
                'event': 'dict',
                'details': {},
                'event_type': 'resolve',
                'description': 'awslimitchecker in myAcct rname found '
                               'no problems'
            }),
            call(self.cls, 'wKey', {
                'event': 'dict',
                'details': {},
                'event_type': 'resolve',
                'description': 'awslimitchecker in myAcct rname found '
                               'no problems'
            })
        ]

    def test_no_account_alias(self):
        self.cls._service_key_crit = 'cKey'
        self.cls._service_key_warn = 'wKey'
        with patch('%s._event_dict' % pb, autospec=True) as m_ed:
            m_ed.return_value = {'event': 'dict', 'details': {}}
            with patch('%s._send_event' % pb, autospec=True) as m_send:
                self.cls.on_success(duration=12.3)
        assert m_ed.mock_calls == [call(self.cls)]
        assert m_send.mock_calls == [
            call(self.cls, 'cKey', {
                'event': 'dict',
                'details': {
                    'duration_seconds': 12.3
                },
                'event_type': 'resolve',
                'description': 'awslimitchecker in rname found '
                               'no problems; run completed in '
                               '12.30 seconds'
            }),
            call(self.cls, 'wKey', {
                'event': 'dict',
                'details': {
                    'duration_seconds': 12.3
                },
                'event_type': 'resolve',
                'description': 'awslimitchecker in rname found '
                               'no problems; run completed in '
                               '12.30 seconds'
            })
        ]

    def test_no_service_key_warn(self):
        self.cls._service_key_crit = 'cKey'
        self.cls._account_alias = 'myAcct'
        with patch('%s._event_dict' % pb, autospec=True) as m_ed:
            m_ed.return_value = {'event': 'dict', 'details': {}}
            with patch('%s._send_event' % pb, autospec=True) as m_send:
                self.cls.on_success(duration=12.3)
        assert m_ed.mock_calls == [call(self.cls)]
        assert m_send.mock_calls == [
            call(self.cls, 'cKey', {
                'event': 'dict',
                'details': {
                    'duration_seconds': 12.3
                },
                'event_type': 'resolve',
                'description': 'awslimitchecker in myAcct rname found '
                               'no problems; run completed in '
                               '12.30 seconds'
            })
        ]


class TestProblemsDict(PagerDutyV1Tester):

    def test_happy_path(self):
        mock_l1 = Mock()
        type(mock_l1).name = 'l1'
        mock_l1.get_warnings.return_value = ['w1']
        mock_l1.get_criticals.return_value = []
        mock_l2 = Mock()
        type(mock_l2).name = 'l2'
        mock_l2.get_warnings.return_value = []
        mock_l2.get_criticals.return_value = []
        mock_l3 = Mock()
        type(mock_l3).name = 'l3'
        mock_l3.get_warnings.return_value = ['w2', 'w3']
        mock_l3.get_criticals.return_value = ['c1', 'c2']
        problems = {
            'S1': {
                'l1': mock_l1,
                'l2': mock_l2
            },
            'S2': {
                'l3': mock_l3
            }
        }
        expected = {
            'S1': {
                'l1': 'S1/l1',
                'l2': 'S1/l2'
            },
            'S2': {
                'l3': 'S2/l3'
            }
        }

        def se_ist(svc, limit, crits, warns, colorize=False):
            return '', '%s/%s' % (svc, limit.name)

        with patch('%s.issue_string_tuple' % pbm) as m_ist:
            m_ist.side_effect = se_ist
            wc, cc, d = self.cls._problems_dict(problems)
        assert wc == 3
        assert cc == 2
        assert d == expected
        assert m_ist.mock_calls == [
            call('S1', mock_l1, [], ['w1'], colorize=False),
            call('S1', mock_l2, [], [], colorize=False),
            call(
                'S2', mock_l3, ['c1', 'c2'], ['w2', 'w3'],
                colorize=False
            )
        ]


class TestOnCritical(PagerDutyV1Tester):

    def test_criticals(self):
        self.cls._account_alias = 'aAlias'
        self.cls._service_key_crit = 'cKey'
        data = {'event': 'data', 'details': {}}
        expected = {
            'event': 'data',
            'event_type': 'trigger',
            'description': 'awslimitchecker in aAlias rname ran in '
                           '12.35 seconds and crossed 2 CRITICAL '
                           'thresholds and 1 WARNING thresholds',
            'details': {
                'duration_seconds': 12.34567,
                'limits': {'foo': 'bar'}
            }
        }
        with patch.multiple(
            pb,
            autospec=True,
            _event_dict=DEFAULT,
            _problems_dict=DEFAULT,
            _send_event=DEFAULT
        ) as mocks:
            mocks['_event_dict'].return_value = data
            mocks['_problems_dict'].return_value = (
                1, 2, {'foo': 'bar'}
            )
            self.cls.on_critical({'p': 'd'}, '', duration=12.34567)
        assert mocks['_event_dict'].mock_calls == [call(self.cls)]
        assert mocks['_problems_dict'].mock_calls == [
            call(self.cls, {'p': 'd'})
        ]
        assert mocks['_send_event'].mock_calls == [
            call(self.cls, 'cKey', expected)
        ]

    def test_no_warnings_no_duration(self):
        self.cls._account_alias = 'aAlias'
        self.cls._service_key_crit = 'cKey'
        data = {'event': 'data', 'details': {}}
        expected = {
            'event': 'data',
            'event_type': 'trigger',
            'description': 'awslimitchecker in aAlias rname crossed '
                           '2 CRITICAL thresholds',
            'details': {
                'limits': {'foo': 'bar'}
            }
        }
        with patch.multiple(
            pb,
            autospec=True,
            _event_dict=DEFAULT,
            _problems_dict=DEFAULT,
            _send_event=DEFAULT
        ) as mocks:
            mocks['_event_dict'].return_value = data
            mocks['_problems_dict'].return_value = (
                0, 2, {'foo': 'bar'}
            )
            self.cls.on_critical({'p': 'd'}, '')
        assert mocks['_event_dict'].mock_calls == [call(self.cls)]
        assert mocks['_problems_dict'].mock_calls == [
            call(self.cls, {'p': 'd'})
        ]
        assert mocks['_send_event'].mock_calls == [
            call(self.cls, 'cKey', expected)
        ]

    def test_exception_no_account_alias(self):
        self.cls._service_key_crit = 'cKey'
        data = {'event': 'data', 'details': {}}
        exc = RuntimeError('foo')
        expected = {
            'event': 'data',
            'event_type': 'trigger',
            'description': 'awslimitchecker in rname ran in '
                           '12.30 seconds and failed with an '
                           'exception: %s' % exc.__repr__(),
            'details': {
                'duration_seconds': 12.3,
                'exception': exc.__repr__()}
        }
        with patch.multiple(
            pb,
            autospec=True,
            _event_dict=DEFAULT,
            _problems_dict=DEFAULT,
            _send_event=DEFAULT
        ) as mocks:
            mocks['_event_dict'].return_value = data
            mocks['_problems_dict'].return_value = (
                1, 2, {'foo': 'bar'}
            )
            self.cls.on_critical(None, None, exc=exc, duration=12.3)
        assert mocks['_event_dict'].mock_calls == [call(self.cls)]
        assert mocks['_problems_dict'].mock_calls == []
        assert mocks['_send_event'].mock_calls == [
            call(self.cls, 'cKey', expected)
        ]


class TestOnWarning(PagerDutyV1Tester):

    def test_duration_alias(self):
        self.cls._account_alias = 'aAlias'
        self.cls._service_key_warn = 'wKey'
        data = {'event': 'data', 'details': {}}
        expected = {
            'event': 'data',
            'event_type': 'trigger',
            'description': 'awslimitchecker in aAlias rname ran in '
                           '12.35 seconds and crossed 1 WARNING '
                           'thresholds',
            'details': {
                'duration_seconds': 12.34567,
                'limits': {'foo': 'bar'}
            }
        }
        with patch.multiple(
            pb,
            autospec=True,
            _event_dict=DEFAULT,
            _problems_dict=DEFAULT,
            _send_event=DEFAULT
        ) as mocks:
            mocks['_event_dict'].return_value = data
            mocks['_problems_dict'].return_value = (
                1, 0, {'foo': 'bar'}
            )
            self.cls.on_warning({'p': 'd'}, '', duration=12.34567)
        assert mocks['_event_dict'].mock_calls == [call(self.cls)]
        assert mocks['_problems_dict'].mock_calls == [
            call(self.cls, {'p': 'd'})
        ]
        assert mocks['_send_event'].mock_calls == [
            call(self.cls, 'wKey', expected)
        ]

    def test_no_duration_no_alias(self):
        self.cls._service_key_warn = 'wKey'
        data = {'event': 'data', 'details': {}}
        expected = {
            'event': 'data',
            'event_type': 'trigger',
            'description': 'awslimitchecker in rname crossed '
                           '1 WARNING thresholds',
            'details': {
                'limits': {'foo': 'bar'}
            }
        }
        with patch.multiple(
            pb,
            autospec=True,
            _event_dict=DEFAULT,
            _problems_dict=DEFAULT,
            _send_event=DEFAULT
        ) as mocks:
            mocks['_event_dict'].return_value = data
            mocks['_problems_dict'].return_value = (
                1, 0, {'foo': 'bar'}
            )
            self.cls.on_warning({'p': 'd'}, '')
        assert mocks['_event_dict'].mock_calls == [call(self.cls)]
        assert mocks['_problems_dict'].mock_calls == [
            call(self.cls, {'p': 'd'})
        ]
        assert mocks['_send_event'].mock_calls == [
            call(self.cls, 'wKey', expected)
        ]



"""
awslimitchecker/services/base.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import abc
import logging
import boto3
from datetime import datetime, timedelta
from awslimitchecker.connectable import Connectable

logger = logging.getLogger(__name__)


class _AwsService(Connectable):
    __metaclass__ = abc.ABCMeta

    #: awslimitchecker's name for the service
    service_name = 'baseclass'

    #: the AWS API name for the service
    api_name = 'baseclass'

    #: the service code for Service Quotas, or None
    quotas_service_code = None

    def __init__(self, warning_threshold, critical_threshold,
                 boto_connection_kwargs, quotas_client):
        """
        Describes an AWS service and its limits, and provides methods to
        query current utilization.

        Constructors of _AwsService subclasses *must not* make any external
        connections; these must be made lazily as needed in other methods.
        _AwsService subclasses should be usable without any external network
        connections.

        :param warning_threshold: the default warning threshold, as an
          integer percentage, for any limits without a specifically-set
          threshold.
        :type warning_threshold: int
        :param critical_threshold: the default critical threshold, as an
          integer percentage, for any limits without a specifically-set
          threshold.
        :type critical_threshold: int
        :param boto_connection_kwargs: Dictionary of keyword arguments to
          pass to boto connection methods.
        :type boto_connection_kwargs: dict
        :param quotas_client: Instance of ServiceQuotasClient
        :type quotas_client: ``ServiceQuotasClient`` or ``None``
        """
        self.warning_threshold = warning_threshold
        self.critical_threshold = critical_threshold
        self._boto3_connection_kwargs = boto_connection_kwargs
        self._quotas_client = quotas_client
        self.conn = None
        self.resource_conn = None
        self.limits = {}
        self.limits = self.get_limits()
        self._have_usage = False
        self._current_account_id = None
        self._cloudwatch_client = None

    @property
    def current_account_id(self):
        """
        Return the numeric Account ID for the account that we are currently
        running against.

        :return: current account ID
        :rtype: str
        """
        if self._current_account_id is not None:
            return self._current_account_id
        kwargs = dict(self._boto3_connection_kwargs)
        sts = boto3.client('sts', **kwargs)
        logger.info(
            "Connected to STS in region %s", sts._client_config.region_name
        )
        cid = sts.get_caller_identity()
        self._current_account_id = cid['Account']
        return cid['Account']

    @abc.abstractmethod
    def find_usage(self):
        """
        Determine the current usage for each limit of this service,
        and update the ``current_usage`` property of each corresponding
        :py:class:`~.AwsLimit` instance.

        This method MUST set ``self._have_usage = True``.

        If the boto3 method being called returns a dict response that can
        include 'NextToken' or another pagination marker, it should be called
        through
        :py:func:`~awslimitchecker.utils.paginate_dict` with the appropriate
        parameters.
        """
        """
        logger.debug("Checking usage for service {n}".format(
            n=self.service_name))
        self.connect()
        usage = self.conn.method_to_get_usage()
        # or, if it needs to be paginated, something like:
        usage = paginate_dict(
            self.conn.method_to_get_usage,
            alc_marker_path=['NextToken'],
            alc_data_path=['ResourceListName'],
            alc_marker_param='NextToken'
        )
        logger.debug("Done checking usage.")
        self._have_usage = True
        """
        raise NotImplementedError('abstract base class')

    @abc.abstractmethod
    def get_limits(self):
        """
        Return all known limits for this service, as a dict of their names
        to :py:class:`~.AwsLimit` objects.

        All limits must have ``self.warning_threshold`` and
        ``self.critical_threshold`` passed into them.

        :returns: dict of limit names to :py:class:`~.AwsLimit` objects
        :rtype: dict
        """
        """
        if self.limits != []:
            return self.limits
        # else define the limits
        """
        raise NotImplementedError('abstract base class')

    @abc.abstractmethod
    def required_iam_permissions(self):
        """
        Return a list of IAM Actions required for this Service to function
        properly. All Actions will be shown with an Effect of "Allow"
        and a Resource of "*".

        :returns: list of IAM Action strings
        :rtype: list
        """
        raise NotImplementedError('abstract base class')

    def set_limit_override(self, limit_name, value, override_ta=True):
        """
        Set a new limit ``value`` for the specified limit, overriding
        the default. If ``override_ta`` is True, also use this value
        instead of any found by Trusted Advisor. This method simply
        passes the data through to the
        :py:meth:`~awslimitchecker.limit.AwsLimit.set_limit_override`
        method of the underlying :py:class:`~.AwsLimit` instance.

        :param limit_name: the name of the limit to override the value for
        :type limit_name: str
        :param value: the new value to set for the limit
        :type value: int
        :param override_ta: whether or not to also override Trusted
          Advisor information
        :type override_ta: bool
        :raises: ValueError if limit_name is not known to this service
        """
        try:
            self.limits[limit_name].set_limit_override(
                value,
                override_ta=override_ta
            )
            logger.debug(
                "Overriding %s limit %s; default=%d override=%d",
                self.service_name,
                limit_name,
                value,
                self.limits[limit_name].default_limit,
            )
        except KeyError:
            raise ValueError("{s} service has no '{l}' limit".format(
                s=self.service_name,
                l=limit_name))

    def _set_ta_limit(self, limit_name, value):
        """
        Set the value for the limit as reported by Trusted Advisor,
        for the specified limit.

        This method should only be called by :py:class:`~.TrustedAdvisor`.

        :param limit_name: the name of the limit to override the value for
        :type limit_name: str
        :param value: the Trusted Advisor limit value
        :type value: int
        :raises: ValueError if limit_name is not known to this service
        """
        try:
            self.limits[limit_name]._set_ta_limit(value)
            logger.debug(
                "Setting %s limit %s TA limit to %d",
                self.service_name,
                limit_name,
                value,
            )
        except KeyError:
            raise ValueError("{s} service has no '{l}' limit".format(
                s=self.service_name,
                l=limit_name))

    def set_threshold_override(self, limit_name, warn_percent=None,
                               warn_count=None, crit_percent=None,
                               crit_count=None):
        """
        Override the default warning and critical thresholds used to evaluate
        the specified limit's usage. Theresholds can be specified as a
        percentage of the limit, or as a usage count, or both.

        :param warn_percent: new warning threshold, percentage used
        :type warn_percent: int
        :param warn_count: new warning threshold, actual count/number
        :type warn_count: int
        :param crit_percent: new critical threshold, percentage used
        :type crit_percent: int
        :param crit_count: new critical threshold, actual count/number
        :type crit_count: int
        """
        try:
            self.limits[limit_name].set_threshold_override(
                warn_percent=warn_percent,
                warn_count=warn_count,
                crit_percent=crit_percent,
                crit_count=crit_count
            )
        except KeyError:
            raise ValueError("{s} service has no '{l}' limit".format(
                s=self.service_name,
                l=limit_name))

    def check_thresholds(self):
        """
        Checks current usage against configured thresholds for all limits
        for this service.

        :returns: a dict of limit name to :py:class:`~.AwsLimit` instance
          for all limits that crossed one or more of their thresholds.
        :rtype: :py:obj:`dict` of :py:class:`~.AwsLimit`
        """
        if not self._have_usage:
            self.find_usage()
        ret = {}
        for name, limit in self.limits.items():
            if limit.check_thresholds() is False:
                ret[name] = limit
        return ret

    def _update_service_quotas(self):
        """
        Update all limits for this service via the Service Quotas service.
        """
        if self.quotas_service_code is None:
            return
        if self._quotas_client is None:
            return
        logger.debug('Updating service quotas for %s', self.service_name)
        for lname in sorted(self.limits.keys()):
            lim = self.limits[lname]
            val = self._quotas_client.get_quota_value(
                lim.quotas_service_code, lim.quota_name,
                units=lim.quotas_unit, converter=lim.quotas_unit_converter
            )
            if val is not None:
                lim._set_quotas_limit(val)

    def _cloudwatch_connection(self):
        """
        Return a connected CloudWatch client instance. ONLY to be used by
        :py:meth:`_get_cloudwatch_usage_latest`.
        """
        if self._cloudwatch_client is not None:
            return self._cloudwatch_client
        kwargs = dict(self._boto3_connection_kwargs)
        if self._max_retries_config is not None:
            kwargs['config'] = self._max_retries_config
        self._cloudwatch_client = boto3.client('cloudwatch', **kwargs)
        logger.info(
            "Connected to cloudwatch in region %s",
            self._cloudwatch_client._client_config.region_name
        )
        return self._cloudwatch_client

    def _get_cloudwatch_usage_latest(
        self, dimensions, metric_name='ResourceCount', period=60
    ):
        """
        Given some metric dimensions, return the value of the latest data point
        for the ``AWS/Usage`` metric specified.

        :param dimensions: list of dicts; dimensions for the metric
        :type dimensions: list
        :param metric_name: AWS/Usage metric name to get
        :type metric_name: str
        :param period: metric period
        :type period: int
        :return: return the metric value (float or int), or None if it cannot
          be retrieved
        :rtype: ``float, int or None``
        """
        conn = self._cloudwatch_connection()
        kwargs = dict(
            MetricDataQueries=[
                {
                    'Id': 'id',
                    'MetricStat': {
                        'Metric': {
                            'Namespace': 'AWS/Usage',
                            'MetricName': metric_name,
                            'Dimensions': dimensions
                        },
                        'Period': period,
                        'Stat': 'Average'
                    }
                }
            ],
            StartTime=datetime.utcnow() - timedelta(hours=1),
            EndTime=datetime.utcnow(),
            ScanBy='TimestampDescending',
            MaxDatapoints=1
        )
        try:
            logger.debug('Querying CloudWatch GetMetricData: %s', kwargs)
            resp = conn.get_metric_data(**kwargs)
        except Exception as ex:
            logger.error(
                'Error querying CloudWatch GetMetricData for AWS/Usage %s: %s',
                metric_name, ex
            )
            return 0
        results = resp.get('MetricDataResults', [])
        if len(results) < 1 or len(results[0]['Values']) < 1:
            logger.warning(
                'No data points found for AWS/Usage metric %s with dimensions '
                '%s; using value of zero!', metric_name, dimensions
            )
            return 0
        logger.debug(
            'CloudWatch metric query returned value of %s with timestamp %s',
            results[0]['Values'][0], results[0]['Timestamps'][0]
        )
        return results[0]['Values'][0]



"""
awslimitchecker/alerts/dummy.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2019 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from .base import AlertProvider

logger = logging.getLogger(__name__)


class Dummy(AlertProvider):
    """
    Dummy alert provider that just outputs alerts to STDOUT (this will
    effectively duplicate awslimitchecker's CLI output).
    """

    def __init__(self, region_name, **_):
        """
        Initialize an AlertProvider class. This MUST be overridden by
        subclasses. All configuration must be passed as keyword arguments
        to the class constructor (these come from ``--alert-config`` CLI
        arguments). Any dependency imports must be made in the constructor.
        The constructor should do as much as possible to validate configuration.

        :param region_name: the name of the region we're connected to
        :type region_name: str
        """
        super(Dummy, self).__init__(region_name)

    def on_success(self, duration=None):
        """
        Method called when no thresholds were breached, and run completed
        successfully. Should resolve any open incidents (if the service supports
        that functionality) or else simply return.

        :param duration: duration of the usage/threshold checking run
        :type duration: float
        """
        print('awslimitchecker in %s found no problems' % self._region_name)
        if duration:
            print('awslimitchecker run duration: %s' % duration)

    def on_critical(self, problems, problem_str, exc=None, duration=None):
        """
        Method called when the run encountered errors, or at least one critical
        threshold was met or crossed.

        :param problems: dict of service name to nested dict of limit name to
          limit, same format as the return value of
          :py:meth:`~.AwsLimitChecker.check_thresholds`. ``None`` if ``exc`` is
          specified.
        :type problems: dict or None
        :param problem_str: String representation of ``problems``, as displayed
          in ``awslimitchecker`` command line output. ``None`` if ``exc`` is
          specified.
        :type problem_str: str or None
        :param exc: Exception object that was raised during the run (optional)
        :type exc: Exception
        :param duration: duration of the run
        :type duration: float
        """
        if exc is not None:
            print('awslimitchecker in %s failed with exception: %s' % (
                self._region_name, exc.__repr__()
            ))
        else:
            print('awslimitchecker in %s found CRITICALS:' % self._region_name)
            print(problem_str)
        if duration:
            print('awslimitchecker run duration: %s' % duration)

    def on_warning(self, problems, problem_str, duration=None):
        """
        Method called when one or more warning thresholds were crossed, but no
        criticals and the run did not encounter any errors.

        :param problems: dict of service name to nested dict of limit name to
          limit, same format as the return value of
          :py:meth:`~.AwsLimitChecker.check_thresholds`.
        :type problems: dict or None
        :param problem_str: String representation of ``problems``, as displayed
          in ``awslimitchecker`` command line output.
        :type problem_str: str or None
        :param duration: duration of the run
        :type duration: float
        """
        print('awslimitchecker in %s found WARNINGS:' % self._region_name)
        print(problem_str)
        if duration:
            print('awslimitchecker run duration: %s' % duration)



"""
awslimitchecker/tests/services/test_ses.py

The latest version of this package is available at:
<https://github.com/jantman/awslimitchecker>

################################################################################
Copyright 2015-2018 Jason Antman <jason@jasonantman.com>

    This file is part of awslimitchecker, also known as awslimitchecker.

    awslimitchecker is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    awslimitchecker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with awslimitchecker.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/awslimitchecker> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
import pytest
from awslimitchecker.services.ses import _SesService
from botocore.exceptions import (
    ClientError, EndpointConnectionError, ConnectTimeoutError
)

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import patch, call, Mock
else:
    from unittest.mock import patch, call, Mock


pbm = 'awslimitchecker.services.ses'  # module patch base
pb = '%s._SesService' % pbm  # class patch pase


class Test_SesService(object):

    def test_init(self):
        """test __init__()"""
        cls = _SesService(21, 43, {}, None)
        assert cls.service_name == 'SES'
        assert cls.api_name == 'ses'
        assert cls.conn is None
        assert cls.warning_threshold == 21
        assert cls.critical_threshold == 43

    def test_get_limits(self):
        cls = _SesService(21, 43, {}, None)
        cls.limits = {}
        res = cls.get_limits()
        assert sorted(res.keys()) == sorted([
            'Daily sending quota',
        ])
        limit = cls.limits['Daily sending quota']
        assert limit.service == cls
        assert limit.def_warning_threshold == 21
        assert limit.def_critical_threshold == 43
        assert limit.default_limit == 200

    def test_get_limits_again(self):
        """test that existing limits dict is returned on subsequent calls"""
        mock_limits = Mock()
        cls = _SesService(21, 43, {}, None)
        cls.limits = mock_limits
        res = cls.get_limits()
        assert res == mock_limits

    def test_find_usage(self):
        mock_conn = Mock()
        mock_conn.get_send_quota.return_value = {
            'Max24HourSend': 123.0,
            'MaxSendRate': 12.0,
            'SentLast24Hours': 122.0
        }

        with patch('%s.connect' % pb) as mock_connect:
            cls = _SesService(21, 43, {}, None)
            cls.conn = mock_conn
            assert cls._have_usage is False
            cls.find_usage()
        assert mock_connect.mock_calls == [call()]
        assert cls._have_usage is True
        assert mock_conn.mock_calls == [call.get_send_quota()]
        assert len(cls.limits['Daily sending quota'].get_current_usage()) == 1
        assert cls.limits[
            'Daily sending quota'
        ].get_current_usage()[0].get_value() == 122.0

    def test_find_usage_invalid_region(self):
        def se_get():
            raise EndpointConnectionError(endpoint_url='myurl')

        mock_conn = Mock()
        mock_conn.get_send_quota.side_effect = se_get

        with patch('%s.connect' % pb) as mock_connect:
            with patch('%s.logger' % pbm) as mock_logger:
                cls = _SesService(21, 43, {}, None)
                cls.conn = mock_conn
                assert cls._have_usage is False
                cls.find_usage()
        assert mock_connect.mock_calls == [call()]
        assert cls._have_usage is False
        assert mock_logger.mock_calls == [
            call.debug('Checking usage for service %s', 'SES'),
            call.warning(
                'Skipping SES: %s',
                'Could not connect to the endpoint URL: "myurl"'
            )
        ]
        assert mock_conn.mock_calls == [call.get_send_quota()]
        assert len(cls.limits['Daily sending quota'].get_current_usage()) == 0

    def test_find_usage_invalid_region_client_error(self):
        resp = {
            'ResponseMetadata': {
                'HTTPStatusCode': 400,
                'RequestId': '7d74c6f0-c789-11e5-82fe-a96cdaa6d564'
            },
            'Error': {
                'Message': 'Unknown',
                'Code': 'AccessDenied',
                'Type': 'Sender'
            }
        }
        ce = ClientError(resp, 'GetSendQuota')

        def se_get():
            raise ce

        mock_conn = Mock()
        mock_conn.get_send_quota.side_effect = se_get

        with patch('%s.connect' % pb) as mock_connect:
            with patch('%s.logger' % pbm) as mock_logger:
                cls = _SesService(21, 43, {}, None)
                cls.conn = mock_conn
                assert cls._have_usage is False
                cls.find_usage()
        assert mock_connect.mock_calls == [call()]
        assert cls._have_usage is False
        assert mock_logger.mock_calls == [
            call.debug('Checking usage for service %s', 'SES'),
            call.warning(
                'Skipping SES: %s', ce
            )
        ]
        assert mock_conn.mock_calls == [call.get_send_quota()]
        assert len(cls.limits['Daily sending quota'].get_current_usage()) == 0

    def test_find_usage_invalid_region_connect_timeout(self):
        ce = ConnectTimeoutError(endpoint_url='http://example.com')

        def se_get():
            raise ce

        mock_conn = Mock()
        mock_conn.get_send_quota.side_effect = se_get

        with patch('%s.connect' % pb) as mock_connect:
            with patch('%s.logger' % pbm) as mock_logger:
                cls = _SesService(21, 43, {}, None)
                cls.conn = mock_conn
                assert cls._have_usage is False
                cls.find_usage()
        assert mock_connect.mock_calls == [call()]
        assert cls._have_usage is False
        assert mock_logger.mock_calls == [
            call.debug('Checking usage for service %s', 'SES'),
            call.warning('Skipping SES: %s', str(ce))
        ]
        assert mock_conn.mock_calls == [call.get_send_quota()]
        assert len(cls.limits['Daily sending quota'].get_current_usage()) == 0

    def test_find_usage_invalid_region_503(self):
        resp = {
            'ResponseMetadata': {
                'HTTPStatusCode': 503,
                'RequestId': '7d74c6f0-c789-11e5-82fe-a96cdaa6d564'
            },
            'Error': {
                'Message': 'Service Unavailable',
                'Code': '503'
            }
        }
        ce = ClientError(resp, 'GetSendQuota')

        def se_get():
            raise ce

        mock_conn = Mock()
        mock_conn.get_send_quota.side_effect = se_get

        with patch('%s.connect' % pb) as mock_connect:
            with patch('%s.logger' % pbm) as mock_logger:
                cls = _SesService(21, 43, {}, None)
                cls.conn = mock_conn
                assert cls._have_usage is False
                cls.find_usage()
        assert mock_connect.mock_calls == [call()]
        assert cls._have_usage is False
        assert mock_logger.mock_calls == [
            call.debug('Checking usage for service %s', 'SES'),
            call.warning(
                'Skipping SES: %s', ce
            )
        ]
        assert mock_conn.mock_calls == [call.get_send_quota()]
        assert len(cls.limits['Daily sending quota'].get_current_usage()) == 0

    def test_find_usage_other_client_error(self):
        resp = {
            'ResponseMetadata': {
                'HTTPStatusCode': 400,
                'RequestId': '7d74c6f0-c789-11e5-82fe-a96cdaa6d564'
            },
            'Error': {
                'Message': 'Not Unknown',
                'Code': 'NotAccessDenied',
                'Type': 'Sender'
            }
        }
        ce = ClientError(resp, 'operation')

        def se_get():
            raise ce

        mock_conn = Mock()
        mock_conn.get_send_quota.side_effect = se_get

        with patch('%s.connect' % pb) as mock_connect:
            with patch('%s.logger' % pbm) as mock_logger:
                cls = _SesService(21, 43, {}, None)
                cls.conn = mock_conn
                assert cls._have_usage is False
                with pytest.raises(ClientError):
                    cls.find_usage()
        assert mock_connect.mock_calls == [call()]
        assert cls._have_usage is False
        assert mock_logger.mock_calls == [
            call.debug('Checking usage for service %s', 'SES')
        ]
        assert mock_conn.mock_calls == [call.get_send_quota()]
        assert len(cls.limits['Daily sending quota'].get_current_usage()) == 0

    def test_update_limits_from_api(self):
        mock_conn = Mock()
        mock_conn.get_send_quota.return_value = {
            'Max24HourSend': 123.0,
            'MaxSendRate': 12.0,
            'SentLast24Hours': 122.0
        }

        with patch('%s.connect' % pb) as mock_connect:
            cls = _SesService(21, 43, {}, None)
            cls.conn = mock_conn
            cls._update_limits_from_api()
        assert mock_connect.mock_calls == [call()]
        assert mock_conn.mock_calls == [call.get_send_quota()]
        assert cls.limits['Daily sending quota'].api_limit == 123.0

    def test_update_limits_from_api_invalid_region(self):
        def se_get():
            raise EndpointConnectionError(endpoint_url='myurl')

        mock_conn = Mock()
        mock_conn.get_send_quota.side_effect = se_get

        with patch('%s.connect' % pb) as mock_connect:
            with patch('%s.logger' % pbm) as mock_logger:
                cls = _SesService(21, 43, {}, None)
                cls.conn = mock_conn
                cls._update_limits_from_api()
        assert mock_connect.mock_calls == [call()]
        assert mock_conn.mock_calls == [call.get_send_quota()]
        assert mock_logger.mock_calls == [
            call.warning(
                'Skipping SES: %s',
                'Could not connect to the endpoint URL: "myurl"'
            )
        ]
        assert cls.limits['Daily sending quota'].api_limit is None

    def test_update_limits_from_api_invalid_region_client_error(self):
        resp = {
            'ResponseMetadata': {
                'HTTPStatusCode': 400,
                'RequestId': '7d74c6f0-c789-11e5-82fe-a96cdaa6d564'
            },
            'Error': {
                'Message': 'Unknown',
                'Code': 'AccessDenied',
                'Type': 'Sender'
            }
        }
        ce = ClientError(resp, 'GetSendQuota')

        def se_get():
            raise ce

        mock_conn = Mock()
        mock_conn.get_send_quota.side_effect = se_get

        with patch('%s.connect' % pb) as mock_connect:
            with patch('%s.logger' % pbm) as mock_logger:
                cls = _SesService(21, 43, {}, None)
                cls.conn = mock_conn
                cls._update_limits_from_api()
        assert mock_connect.mock_calls == [call()]
        assert mock_conn.mock_calls == [call.get_send_quota()]
        assert mock_logger.mock_calls == [
            call.warning('Skipping SES: %s', ce)
        ]
        assert cls.limits['Daily sending quota'].api_limit is None

    def test_update_limits_from_api_invalid_region_connect_timeout(self):
        ce = ConnectTimeoutError(endpoint_url='http://example.com')

        def se_get():
            raise ce

        mock_conn = Mock()
        mock_conn.get_send_quota.side_effect = se_get

        with patch('%s.connect' % pb) as mock_connect:
            with patch('%s.logger' % pbm) as mock_logger:
                cls = _SesService(21, 43, {}, None)
                cls.conn = mock_conn
                cls._update_limits_from_api()
        assert mock_connect.mock_calls == [call()]
        assert mock_conn.mock_calls == [call.get_send_quota()]
        assert mock_logger.mock_calls == [
            call.warning('Skipping SES: %s', str(ce))
        ]
        assert cls.limits['Daily sending quota'].api_limit is None

    def test_update_limits_from_api_invalid_region_503(self):
        resp = {
            'ResponseMetadata': {
                'HTTPStatusCode': 503,
                'RequestId': '7d74c6f0-c789-11e5-82fe-a96cdaa6d564'
            },
            'Error': {
                'Message': 'Service Unavailable',
                'Code': '503'
            }
        }
        ce = ClientError(resp, 'GetSendQuota')

        def se_get():
            raise ce

        mock_conn = Mock()
        mock_conn.get_send_quota.side_effect = se_get

        with patch('%s.connect' % pb) as mock_connect:
            with patch('%s.logger' % pbm) as mock_logger:
                cls = _SesService(21, 43, {}, None)
                cls.conn = mock_conn
                cls._update_limits_from_api()
        assert mock_connect.mock_calls == [call()]
        assert mock_conn.mock_calls == [call.get_send_quota()]
        assert mock_logger.mock_calls == [
            call.warning('Skipping SES: %s', ce)
        ]
        assert cls.limits['Daily sending quota'].api_limit is None

    def test_update_limits_from_api_other_client_error(self):
        resp = {
            'ResponseMetadata': {
                'HTTPStatusCode': 400,
                'RequestId': '7d74c6f0-c789-11e5-82fe-a96cdaa6d564'
            },
            'Error': {
                'Message': 'Not Unknown',
                'Code': 'NotAccessDenied',
                'Type': 'Sender'
            }
        }
        ce = ClientError(resp, 'GetSendQuota')

        def se_get():
            raise ce

        mock_conn = Mock()
        mock_conn.get_send_quota.side_effect = se_get

        with patch('%s.connect' % pb) as mock_connect:
            with patch('%s.logger' % pbm) as mock_logger:
                cls = _SesService(21, 43, {}, None)
                cls.conn = mock_conn
                with pytest.raises(ClientError):
                    cls._update_limits_from_api()
        assert mock_connect.mock_calls == [call()]
        assert mock_conn.mock_calls == [call.get_send_quota()]
        assert mock_logger.mock_calls == []
        assert cls.limits['Daily sending quota'].api_limit is None

    def test_required_iam_permissions(self):
        cls = _SesService(21, 43, {}, None)
        assert cls.required_iam_permissions() == [
            'ses:GetSendQuota'
        ]
